unit uSaveForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Registry;

type
	TWinProcEvent = procedure(var Message: TMessage;
  	var CallDef: Boolean) of object;

	TWindowHook = class;

  TSaveForm = class(TComponent)
  private
    { Private declarations }
		FRegKey: String;
    RestoreToMax: Boolean;
    WinHook: TWindowHook;
    OldRestoreEvent: TNotifyEvent;
    function Form: TForm;
    function GetEnabled: Boolean;
    procedure SetEnabled(NewCond: Boolean);
		procedure DoRestore(Sender: TObject);
		procedure DoWinHook(var Message: TMessage; var CallDef: Boolean);
  protected
    { Protected declarations }
    procedure Loaded; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
		procedure SaveState;
  	procedure RestoreState;
  published
    { Published declarations }
    property Enabled: Boolean
    	read GetEnabled write SetEnabled default FALSE;
		property RegistryKey: string
			read FRegKey write FRegKey;
  end;

	TWindowHook = class(TPersistent)
  private
  	FControl: TWinControl;
    FHandle: THandle;
    FEnabled, FWorkInDesign: Boolean;
    FOnWndProc: TWinProcEvent;
    FOldProc: TFarProc;
    FNewProc: Pointer;
    procedure HookProc(var Message: TMessage);
    procedure SetEnabled(NewCond: Boolean);
    procedure SetWorkInDesign(NewCond: Boolean);
  protected
  public
    constructor Create(AControl: TWinControl);
		destructor Destroy; override;
    procedure	Hook;
    procedure UnHook;
		procedure DoDefault(var Message: TMessage);
    property Control: TWinControl read FControl;
    property Handle: THandle read FHandle;
    property OldProc: TFarProc read FOldProc;
    property NewProc: Pointer read FNewProc;
  published
  	property Enabled: Boolean
    	read FEnabled write SetEnabled default FALSE;
    property WorkInDesign: Boolean
    	read FWorkInDesign write SetWorkInDesign default FALSE;
  	property OnWndProc: TWinProcEvent
    	read FOnWndProc write FOnWndProc;
  end;

function GetWinPos(Wnd: HWnd): TRect;
function GetWinShowState(Wnd: HWnd): Integer;
function GetWinPlacement(Wnd: HWnd): TWindowPlacement;

procedure Register;

implementation

// Window placement methods
function GetWinPos(Wnd: HWnd): TRect;
begin
	Result := GetWinPlacement(Wnd).rcNormalPosition;
  with Result do begin
  	Dec(Right, Left);
    Dec(Bottom, Top);
  end;
end;

function GetWinShowState(Wnd: HWnd): Integer;
begin
  Result := GetWinPlacement(Wnd).ShowCmd;
end;

function GetWinPlacement(Wnd: HWnd): TWindowPlacement;
begin
	FillChar(Result, SizeOf(Result), 0);
	Result.Length := SizeOf(Result);
	GetWindowPlacement(Wnd, @Result);
end;

// TSaveForm
constructor TSaveForm.Create(AOwner : TComponent);
begin
	inherited Create(AOwner);
 	FRegKey := 'Software\Company\Application\Form';
  WinHook := TWindowHook.Create(AOwner as TForm);
  WinHook.OnWndProc := DoWinHook;
end;

destructor TSaveForm.Destroy;
begin
	WinHook.Free;
	inherited Destroy;
end;

procedure TSaveForm.Loaded;
begin
  inherited Loaded;
  if Enabled and not (csDesigning in ComponentState) then
  	RestoreState;
end;

function TSaveForm.Form: TForm;
begin
	Result := WinHook.Control as TForm;
end;

function TSaveForm.GetEnabled: Boolean;
begin
	Result := WinHook.Enabled;
end;

procedure TSaveForm.SetEnabled(NewCond: Boolean);
begin
	WinHook.Enabled := NewCond;
end;

procedure TSaveForm.DoRestore(Sender: TObject);
begin
	with Application, Form do begin
		ShowMainForm := TRUE;
    if RestoreToMax then
    	Perform(WM_SYSCOMMAND, SC_MAXIMIZE, 0)
    else
    	Perform(WM_SYSCOMMAND, SC_RESTORE, 0);
  	Show;
  	OnRestore := OldRestoreEvent;
  	if Assigned(OldRestoreEvent) then
  		OnRestore(Form);
	end;
end;

procedure TSaveForm.DoWinHook(var Message: TMessage; var CallDef: Boolean);
begin
	if message.Msg = WM_DESTROY then
  	SaveState;
end;

procedure TSaveForm.SaveState;
var
	WinPos: TRect;
	Reg: TRegIniFile;
  ShowState: Integer;
begin
	try
		reg := TRegIniFile.Create(FRegKey);
		with Form do begin
    	RestoreToMax := FALSE;
    	winPos := GetWinPos(Handle);
      ShowState := GetWinShowState(Application.Handle);
      if WindowState = wsMaximized then
      	if ShowState = SW_SHOWMINIMIZED	then
      		RestoreToMax := TRUE
        else
        	ShowState := SW_SHOWMAXIMIZED;
      with reg, winPos do begin
      	WriteInteger('', 'Top', Top);
				WriteInteger('', 'Left', Left);
				WriteInteger('', 'Width', Right);
				WriteInteger('', 'Height', Bottom);
				WriteInteger('', 'State', ShowState);
       	WriteBool('', 'RestoreToMax', RestoreToMax);
      end;
		end;
	finally
		reg.Free;
	end;
end;

procedure TSaveForm.RestoreState;
var
	WinPos: TRect;
	Reg: TRegIniFile;
  ShowState: Integer;
  WinPlacement: TWindowPlacement;
begin
	try
		reg := TRegIniFile.Create(FRegKey);
    with winPos, reg do begin
  		Left := ReadInteger('', 'Left', 0);
  		Top := ReadInteger('', 'Top', 0);
  		Right := ReadInteger('', 'Width', 0);
  		Bottom := ReadInteger('', 'Height', 0);
      if (Top <> 0) and (Left <> 0) and
      (Right <> 0) and (Bottom <> 0) then
      	with Application, Form do begin
        	SetBounds(winPos.Left, winPos.Top, Right, Bottom);
        	ShowState := ReadInteger('', 'State', LongInt(wsNormal));
          RestoreToMax := ReadBool('', 'RestoreToMax', FALSE);
          case ShowState of
  					SW_NORMAL: WindowState := wsNormal;
  					SW_SHOWMINIMIZED: begin
            		if RestoreToMax then
            			WindowState := wsMaximized
                else
           				WindowState := wsMinimized;
               	ShowMainForm := FALSE;
               	OldRestoreEvent := OnRestore;
        				OnRestore := DoRestore;
               	Minimize;
           		end;
           	SW_SHOWMAXIMIZED: WindowState := wsMaximized;
          end;
        end;
    end;
	finally
		reg.Free;
	end;
end;

// TWindowHook
constructor TWindowHook.Create(AControl: TWinControl);
begin
	inherited Create;
  FControl := AControl;
  FHandle := FControl.Handle;
  FEnabled := FALSE;
  FWorkInDesign := FALSE;
end;

destructor TWindowHook.Destroy;
begin
	UnHook;
	inherited Destroy;
end;

procedure TWindowHook.SetEnabled(NewCond: Boolean);
begin
	if NewCond <> FEnabled then
  	if NewCond then begin
	    FEnabled := TRUE;
    	Hook
    end else begin
    	UnHook;
	    FEnabled := FALSE;
    end;
end;

procedure TWindowHook.SetWorkInDesign(NewCond: Boolean);
begin
	if NewCond <> FWorkInDesign then
  	if csDesigning in FControl.ComponentState then begin
  		if NewCond then begin
		  	FWorkInDesign := TRUE;
				Hook
      end else begin
				UnHook;
        FWorkInDesign := FALSE;
      end;
		end else
    	FWorkInDesign := NewCond;
end;

procedure	TWindowHook.Hook;
begin
	if FEnabled and (FWorkInDesign or
  not	(csDesigning in FControl.ComponentState)) then begin
  	FHandle := FControl.Handle;
  	FOldProc := TFarProc(GetWindowLong(FHandle, GWL_WNDPROC));
  	FNewProc := MakeObjectInstance(HookProc);
  	SetWindowLong(FHandle, GWL_WNDPROC, LongInt(FNewProc));
  end;
end;

procedure TWindowHook.UnHook;
begin
	if FEnabled and (FWorkInDesign or
  not	(csDesigning in FControl.ComponentState)) then begin
	  SetWindowLong(FHandle, GWL_WNDPROC, LongInt(FOldProc));
  	if Assigned(FNewProc) then
  		FreeObjectInstance(FNewProc);
  	FNewProc := nil;
  end;
end;

procedure TWindowHook.HookProc(var Message: TMessage);
var
	CallDef: Boolean;
begin
	callDef := TRUE;
	if Assigned(FOnWndProc) then
	  FOnWndProc(message, callDef);
  if callDef then
  	DoDefault(message);
end;

procedure TWindowHook.DoDefault(var Message: TMessage);
begin
	with message do
  	Result := CallWindowProc(FOldProc, FHandle, Msg, wParam, lParam);
end;

procedure Register;
begin
  RegisterComponents('Freeware', [TSaveForm]);
end;

end.
